# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Feature generators module for the Oumi (Open Universal Machine Intelligence) library.

This module provides classes to generate model input features, which can be used
in multiple contexts (datasets, collators, etc).
"""

from oumi.core.feature_generators.base_feature_generator import (
    BaseConversationFeatureGenerator,
    FeatureGeneratorOptions,
)
from oumi.core.feature_generators.vision_language_conversation_feature_generator import (  # noqa: E501
    VisionLanguageConversationFeatureGenerator,
)

__all__ = [
    "BaseConversationFeatureGenerator",
    "FeatureGeneratorOptions",
    "VisionLanguageConversationFeatureGenerator",
]
