# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: conversation.proto
# Protobuf Python Version: 5.27.5
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 5, 27, 5, "", "conversation.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x12\x63onversation.proto\x12\x17oumi.core.conversations"2\n\x08\x44\x61taBlob\x12\x11\n\tmime_type\x18\x01 \x01(\t\x12\x13\n\x0b\x62inary_data\x18\x02 \x01(\x0c"\xe1\x01\n\x0b\x43ontentPart\x12\x37\n\x04type\x18\x01 \x01(\x0e\x32).oumi.core.conversations.ContentPart.Type\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\t\x12/\n\x04\x62lob\x18\x03 \x01(\x0b\x32!.oumi.core.conversations.DataBlob"W\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\x08\n\x04TEXT\x10\x01\x12\x0e\n\nIMAGE_PATH\x10\x02\x12\r\n\tIMAGE_URL\x10\x03\x12\x10\n\x0cIMAGE_BINARY\x10\x04"w\n\x07Message\x12\n\n\x02id\x18\x01 \x01(\t\x12+\n\x04role\x18\x02 \x01(\x0e\x32\x1d.oumi.core.conversations.Role\x12\x33\n\x05parts\x18\x03 \x03(\x0b\x32$.oumi.core.conversations.ContentPart"\xd3\x01\n\x0c\x43onversation\x12\x17\n\x0f\x63onversation_id\x18\x01 \x01(\t\x12\x32\n\x08messages\x18\x02 \x03(\x0b\x32 .oumi.core.conversations.Message\x12\x45\n\x08metadata\x18\x03 \x03(\x0b\x32\x33.oumi.core.conversations.Conversation.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01*K\n\x04Role\x12\x14\n\x10ROLE_UNSPECIFIED\x10\x00\x12\n\n\x06SYSTEM\x10\x01\x12\x08\n\x04USER\x10\x02\x12\r\n\tASSISTANT\x10\x03\x12\x08\n\x04TOOL\x10\x04\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "conversation_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_CONVERSATION_METADATAENTRY"]._loaded_options = None
    _globals["_CONVERSATION_METADATAENTRY"]._serialized_options = b"8\001"
    _globals["_ROLE"]._serialized_start = 662
    _globals["_ROLE"]._serialized_end = 737
    _globals["_DATABLOB"]._serialized_start = 47
    _globals["_DATABLOB"]._serialized_end = 97
    _globals["_CONTENTPART"]._serialized_start = 100
    _globals["_CONTENTPART"]._serialized_end = 325
    _globals["_CONTENTPART_TYPE"]._serialized_start = 238
    _globals["_CONTENTPART_TYPE"]._serialized_end = 325
    _globals["_MESSAGE"]._serialized_start = 327
    _globals["_MESSAGE"]._serialized_end = 446
    _globals["_CONVERSATION"]._serialized_start = 449
    _globals["_CONVERSATION"]._serialized_end = 660
    _globals["_CONVERSATION_METADATAENTRY"]._serialized_start = 613
    _globals["_CONVERSATION_METADATAENTRY"]._serialized_end = 660
# @@protoc_insertion_point(module_scope)
