{% set role_suffix = ': ' %}
{% set turn_suffix = '\n' %}

{% for message in messages %}
    {% if message['role'].upper() == 'ASSISTANT' %}
        {{ message['role'].upper() + role_suffix }}{% generation %}{{message['content'] | trim}}{% endgeneration %}{% if not loop.last %}{{ turn_suffix }}{% endif %}
    {% else %}
        {{ message['role'].upper() + role_suffix + message['content'] | trim }}{% if not loop.last %}{{ turn_suffix }}{% endif %}
    {% endif %}
{% endfor %}
{% if add_generation_prompt %}
    {{ turn_suffix + "ASSISTANT" + role_suffix }}
{% endif %}
